:root {
  font-family:
    Inter, ui-sans-serif, system-ui, -apple-system, BlinkMacSystemFont, 'Segoe UI', sans-serif;
  line-height: 1.5;
  font-weight: 400;
  color: #eff5ff;
  background:
    radial-gradient(circle at top, rgba(37, 99, 235, 0.2), transparent 30%),
    linear-gradient(180deg, #0b1020 0%, #09111f 100%);
  font-synthesis: none;
  text-rendering: optimizeLegibility;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
}

* {
  box-sizing: border-box;
}

html {
  scroll-behavior: smooth;
}

body {
  margin: 0;
  min-width: 320px;
  min-height: 100vh;
}

body,
button,
input,
select {
  font: inherit;
}

button,
select {
  color: inherit;
}

#root {
  min-height: 100vh;
}
