SET search_path TO public;

-- Clear runtime data while leaving schema objects and RPCs in place.
TRUNCATE TABLE
    transactions,
    proof_attempts,
    task_results,
    tasks,
    verified_lemmas,
    research_episodes,
    research_jobs,
    credit_balances,
    nodes
RESTART IDENTITY CASCADE;
