"""Research Capacitor CLI — top-level Typer application."""

from __future__ import annotations

import typer

from rc.commands import (
    balance,
    init,
    ledger,
    research_run,
    research_status,
    rotate_secret,
    serve,
    status,
    submit,
)

app = typer.Typer(
    name="rc",
    help="Research Capacitor — decentralized inference network for mathematical research.",
    no_args_is_help=True,
)

app.command(name="init")(init.run)
app.command(name="serve")(serve.run)
app.command(name="submit")(submit.run)
app.command(name="status")(status.run)
app.command(name="research-run")(research_run.run)
app.command(name="research-status")(research_status.run)
app.command(name="balance")(balance.run)
app.command(name="ledger")(ledger.run)
app.command(name="rotate-secret")(rotate_secret.run)

if __name__ == "__main__":
    app()
